@@ -61,7 +61,7 @@ class PushbulletAsyncIterator(AsyncIterator, Generic[T]):
 
         # Internal management
         self.log = logging.getLogger(__name__ + "." + self.__class__.__name__)
-        self.loop: asyncio.BaseEventLoop = None
+        self.loop: asyncio.AbstractEventLoop = None
         self._paused: bool = False
         self._stop: bool = None  # Instructed to stop or end of iterator
         self._objects: List[Dict] = []  # Use as FIFO queue of objects retrieved
@@ -247,8 +247,8 @@ class AsyncPushbullet(Pushbullet):
     def __init__(self, api_key: str = None, verify_ssl: bool = None, *kargs, **kwargs):
         Pushbullet.__init__(self, api_key, *kargs, **kwargs)
 
-        self.loop: asyncio.BaseEventLoop = None
-        self._aio_session: aiohttp.ClientSession = None
+        self.loop: Optional[asyncio.AbstractEventLoop] = None
+        self._aio_session: Optional[aiohttp.ClientSession] = None
         self.verify_ssl: bool = verify_ssl
 
     async def __aenter__(self):
@@ -982,44 +982,3 @@ class AsyncPushbullet(Pushbullet):
         data = xfer.get("data")
         xfer["msg"] = await self._async_push(data)
         return next(gen)
-
-    # Transfer file sub-function
-    async def async_push_file_simple(self,
-                               file_path: str,
-                               use_transfer_sh: bool = True,
-                               quiet: bool = False,
-                               title: str = None,
-                               body: str = None,
-                               target_device: Device = None):
-        if not os.path.isfile(file_path):
-            # TODO: this error handler is from the command line version
-            print("File not found:", file_path, file=sys.stderr)
-            return
-
-        show_progress = not quiet
-        if use_transfer_sh:
-            if not quiet:
-                print("Uploading file to transfer.sh ... {}".format(file_path))
-
-            info: Dict = await self.async_upload_file_to_transfer_sh(file_path=file_path,
-                                                                     show_progress=show_progress)
-
-        else:
-            if not quiet:
-                print("Uploading file to Pushbullet ... {}".format(file_path))
-            info: Dict = await self.async_upload_file(file_path=file_path,
-                                                      show_progress=show_progress)
-
-        file_url: str = info["file_url"]
-        file_type: str = info["file_type"]
-        file_name: str = info["file_name"]
-        if not quiet:
-            print("Pushing file ... {}".format(file_url))
-
-        title = title or "File: {}".format(file_name)
-        await self.async_push_file(file_name=file_name,
-                                   file_type=file_type,
-                                   file_url=file_url,
-                                   title=title,
-                                   body=body or file_url,
-                                   device=target_device)
