@@ -6,10 +6,12 @@ A command line script for listening for pushes.
 usage: command_line_listen.py [-h] [-k KEY] [--key-file KEY_FILE] [-e]
                               [-x EXEC [EXEC ...]]
                               [-s EXEC_SIMPLE [EXEC_SIMPLE ...]]
+                              [-p EXEC_PYTHON [EXEC_PYTHON ...]] [-t TIMEOUT]
                               [--throttle-count THROTTLE_COUNT]
                               [--throttle-seconds THROTTLE_SECONDS]
                               [-d DEVICE] [--list-devices] [--proxy PROXY]
-                              [--debug] [-v] [--oauth2]
+                              [--debug] [-v] [-q] [--oauth2] [--clear-oauth2]
+                              [--version]
 
 optional arguments:
   -h, --help            show this help message and exit
@@ -31,6 +33,15 @@ optional arguments:
                         script can write lines back to stdout to send a single
                         push back. The first line of stdout will be the title,
                         and subsequent lines will be the body.
+  -p EXEC_PYTHON [EXEC_PYTHON ...], --exec-python EXEC_PYTHON [EXEC_PYTHON ...]
+                        ACTION: Load the given python file and execute it by
+                        calling its on_push(p, pb) function with 2 arguments:
+                        the push that was received and a live/connected
+                        AsyncPushbullet object with which responses may be
+                        sent.
+  -t TIMEOUT, --timeout TIMEOUT
+                        Timeout in seconds to use for actions being called
+                        (default 30).
   --throttle-count THROTTLE_COUNT
                         Pushes will be throttled to this many pushes (default
                         10) in a certain number of seconds (default 10)
@@ -43,25 +54,30 @@ optional arguments:
   --proxy PROXY         Optional web proxy
   --debug               Turn on debug logging
   -v, --verbose         Turn on verbose logging (INFO messages)
+  -q, --quiet           Suppress all output
   --oauth2              Register your command line tool using OAuth2
-
+  --clear-oauth2        Clears/unregisters the oauth2 token
+  --version             show program's version number and exit
 
 """
 import argparse
 import asyncio
+import importlib.util
 import json
 import logging
+import math
 import os
 import pprint
 import sys
 import textwrap
 import threading
 import time
+import traceback
+import types
 from functools import partial
 from typing import List
 
-from asyncpushbullet import AsyncPushbullet
-from asyncpushbullet import Device
+from asyncpushbullet import AsyncPushbullet, __version__
 from asyncpushbullet import InvalidKeyError, PushbulletError
 from asyncpushbullet import LiveStreamListener
 from asyncpushbullet import errors
@@ -72,27 +88,29 @@ __email__ = "rob@iHarder.net"
 
 DEFAULT_THROTTLE_COUNT = 10
 DEFAULT_THROTTLE_SECONDS = 10
+DEFAULT_COMMAND_TIMEOUT = 30
 ENCODING = "utf-8"
 LOG = logging.getLogger(__name__)
 
 
 def main():
+    # sys.argv.append("--clear-oauth2")
     # sys.argv.append("--oauth2")
+    # sys.argv.append("--version")
     # sys.argv.append("-h")
     # sys.argv.append("-v")
     # sys.argv.append("--debug")
     # sys.argv += ["-k", "badkey"]
     # sys.argv += ["--key-file", "../api_key.txt"]
     # sys.argv += ["--device", "Kanga"]
+    # sys.argv += ["--timeout", "3"]
     # sys.argv.append("--echo")
-    # sys.argv += ["--proxy", ""]
+    # sys.argv += ["--proxy", "foo"]
     # sys.argv.append("--list-devices")
-
     # sys.argv += ["--exec-simple", r"C:\windows\System32\clip.exe"]
     # sys.argv += ["--exec", r"C:\windows\System32\notepad.exe"]
-    # sys.argv += ["--exec", r"c:\python37-32\python.exe", r"C:\Users\rharder\Documents\Programming\asyncpushbullet\examples\respond_to_listen_imagesnap.py"]
-    # sys.argv += ["--exec-simple", r"c:\python37-32\python.exe", r"C:\Users\rharder\Documents\Programming\asyncpushbullet\examples\respond_to_listen_exec_simple.py"]
-
+    # sys.argv += ["--exec-python", r"../examples/exec_python_example.py"]
+    # sys.argv += ["--exec-python", r"../examples/exec_python_imagesnap.py"]
     # oauth2.PREFS.set(oauth2.OAUTH2_TOKEN_KEY, None)
 
     args = parse_args()
@@ -108,6 +126,7 @@ def do_main(args):
         exit_code = errors.__ERR_KEYBOARD_INTERRUPT__
     except Exception as ex:
         print("Error:", ex, file=sys.stderr)
+        traceback.print_tb(sys.exc_info()[2])
         exit_code = errors.__ERR_UNKNOWN__
     finally:
         return exit_code or errors.__EXIT_NO_ERROR__
@@ -122,16 +141,23 @@ async def _run(args):
         print("Log level: INFO")
         logging.basicConfig(level=logging.INFO)
 
+    # Clear the oauth2 token?
+    if args.clear_oauth2:
+        oauth2.clear_oauth2_key()
+        print("Successfully cleared/unregistered the oauth2 token.")
+        print("The asyncpushbullet command line tools no longer have access to your pushbullet account.")
+        sys.exit(errors.__EXIT_NO_ERROR__)
+
     # Request setting up oauth2 access?
     if args.oauth2:
         token = await oauth2.async_gain_oauth2_access()
         if token:
             print("Successfully authenticated using OAuth2.")
             print("You should now be able to use the command line tools without specifying an API key.")
-            sys.exit(0)
+            sys.exit(errors.__EXIT_NO_ERROR__)
         else:
             print("There was a problem authenticating.")
-            sys.exit(1)
+            sys.exit(errors.__ERR_UNKNOWN__)
 
     # Find a valid API key
     api_key = try_to_find_key(args, not args.quiet)
@@ -140,24 +166,27 @@ async def _run(args):
         sys.exit(errors.__ERR_API_KEY_NOT_GIVEN__)
 
     # Proxy
-    proxy = args.proxy or os.environ.get("https_proxy") or os.environ.get("http_proxy")
+    proxy = lambda: args.proxy or os.environ.get("https_proxy") or os.environ.get("http_proxy")
 
     # List devices?
     if args.list_devices:
         print("Devices:")
         try:
-            async with AsyncPushbullet(api_key, proxy=proxy) as pb:
+            _proxy = proxy() if callable(proxy) else proxy
+
+            async with AsyncPushbullet(api_key, proxy=_proxy) as pb:
                 async for dev in pb.devices_asynciter():
                     print("\t", dev.nickname)
 
         except InvalidKeyError as exc:
             print(exc, file=sys.stderr)
-            sys.exit(errors.__ERR_INVALID_API_KEY__)
+            return errors.__ERR_INVALID_API_KEY__
         except PushbulletError as exc:
             print(exc, file=sys.stderr)
-            sys.exit(errors.__ERR_CONNECTING_TO_PB__)
+            return errors.__ERR_CONNECTING_TO_PB__
         else:
-            sys.exit(0)
+            # sys.exit(0)
+            return errors.__EXIT_NO_ERROR__
 
     # Throttle
     throttle_count = args.throttle_count
@@ -166,33 +195,39 @@ async def _run(args):
     # Device
     device = args.device
 
+    # Timeout
+    timeout = DEFAULT_COMMAND_TIMEOUT
+    if args.timeout:
+        timeout = float(args.timeout)
+
     # Create ListenApp
     listen_app = ListenApp(api_key,
                            proxy=proxy,
                            throttle_count=throttle_count,
                            throttle_seconds=throttle_seconds,
-                           device=device)
+                           device=device,
+                           timeout=timeout)
 
     # Windows needs special event loop in order to launch processes on it
-    proc_loop = None  # type: asyncio.BaseEventLoop
+    proc_loop: asyncio.BaseEventLoop
     if sys.platform == 'win32':
         proc_loop = asyncio.ProactorEventLoop()
     else:
         proc_loop = asyncio.new_event_loop()  # Processes
         asyncio.get_child_watcher()  # Main loop
 
-    def _run(loop):
+    def _thread_run(loop):
         asyncio.set_event_loop(loop)
         loop.run_forever()
 
-    threading.Thread(target=partial(_run, proc_loop), name="Thread-proc", daemon=True).start()
+    threading.Thread(target=partial(_thread_run, proc_loop), name="Thread-proc", daemon=True).start()
 
     # Add actions from command line arguments
     if args.exec:
         for cmd_opts in args.exec:
             cmd_path = cmd_opts[0]
             cmd_args = cmd_opts[1:]
-            action = ExecutableAction(cmd_path, cmd_args, loop=proc_loop)
+            action = ExecutableAction(cmd_path, cmd_args, loop=proc_loop, timeout=timeout)
             listen_app.add_action(action)
 
     # Add actions from command line arguments
@@ -200,27 +235,35 @@ async def _run(args):
         for cmd_opts in args.exec_simple:
             cmd_path = cmd_opts[0]
             cmd_args = cmd_opts[1:]
-            action = ExecutableActionSimplified(cmd_path, cmd_args, loop=proc_loop)
+            action = ExecutableActionSimplified(cmd_path, cmd_args, loop=proc_loop, timeout=timeout)
             listen_app.add_action(action)
 
+    # Add actions from command line arguments
+    if args.exec_python:
+        for cmd_opts in args.exec_python:
+            cmd_path = cmd_opts[0]
+            action = ExecutableActionPython(cmd_path)
+            listen_app.add_action(action)
+
+    # Echo
+    if args.echo:
+        listen_app.add_action(EchoAction())
+
     # Default action if none specified
-    if not listen_app.actions:
+    if not listen_app._actions:
         print("No actions specified -- defaulting to Echo.")
         listen_app.add_action(EchoAction())
 
-    loop = asyncio.get_event_loop()
-
-    exit_code = None
+    exit_code: int = None
     try:
-        # exit_code = loop.run_until_complete(listen_app.run())
         exit_code = await listen_app.run()
-    except KeyboardInterrupt as e:
+    except KeyboardInterrupt:
         print("Caught keyboard interrupt")
     finally:
-        loop.run_until_complete(listen_app.close())
+        await listen_app.close()
         if exit_code is None:
-            exit_code = 0
-        sys.exit(exit_code)
+            exit_code = errors.__EXIT_NO_ERROR__
+        return exit_code
         # END OF PROGRAM
 
 
@@ -280,7 +323,15 @@ def parse_args():
                         push back.  The first line of stdout will be the title, and
                         subsequent lines will be the body.
                         """))
-
+    parser.add_argument("-p", "--exec-python", nargs="+", action="append",
+                        help=textwrap.dedent("""
+                        ACTION: Load the given python file and execute it by calling
+                        its on_push(p, pb) function with 2 arguments: the push that was
+                        received and a live/connected AsyncPushbullet object with which
+                        responses may be sent.
+                        """))
+    parser.add_argument("-t", "--timeout", help="Timeout in seconds to use for actions being called (default {})."
+                        .format(DEFAULT_COMMAND_TIMEOUT))
     parser.add_argument("--throttle-count", type=int, default=DEFAULT_THROTTLE_COUNT,
                         help=textwrap.dedent("""
                         Pushes will be throttled to this many pushes (default {})
@@ -299,7 +350,10 @@ def parse_args():
     parser.add_argument("--proxy", help="Optional web proxy")
     parser.add_argument("--debug", action="store_true", help="Turn on debug logging")
     parser.add_argument("-v", "--verbose", action="store_true", help="Turn on verbose logging (INFO messages)")
+    parser.add_argument("-q", "--quiet", action="store_true", help="Suppress all output")
     parser.add_argument("--oauth2", action="store_true", help="Register your command line tool using OAuth2")
+    parser.add_argument("--clear-oauth2", action="store_true", help="Clears/unregisters the oauth2 token")
+    parser.add_argument("--version", action="version", version='%(prog)s ' + __version__)
 
     args = parser.parse_args()
 
@@ -316,13 +370,7 @@ class Action:
     def __init__(self):
         self.log = logging.getLogger(__name__ + "." + self.__class__.__name__)
 
-    async def do(self, push: dict, app):
-        """
-
-        :param dict push:
-        :param ListenApp app:
-        :return:
-        """
+    async def on_push(self, push: dict, pb: AsyncPushbullet):
         pass
 
     def __repr__(self):
@@ -332,10 +380,8 @@ class Action:
 class EchoAction(Action):
     """ Echoes pushes in json format to standard out. """
 
-    async def do(self, push: dict, app):  # pb: AsyncPushbullet):  # , device:Device):
-        # json_push = json.dumps(push)
-        # print(json_push, end="\n\n", flush=True)
-        print("Echo push:", pprint.pformat(push))
+    async def on_push(self, push: dict, pb: AsyncPushbullet):
+        print("Echo push: {}".format(pprint.pformat(push)), flush=True)
 
 
 class ExecutableAction(Action):
@@ -374,7 +420,7 @@ class ExecutableAction(Action):
         def process_exited(self):
             print("_ProcessProtocol.process_exited")
 
-    def __init__(self, path_to_executable, args_for_exec=(), loop: asyncio.BaseEventLoop = None, timeout=30):
+    def __init__(self, path_to_executable, args_for_exec=(), loop: asyncio.BaseEventLoop = None, timeout=None):
         super().__init__()
         self.path_to_executable = path_to_executable
         self.args_for_exec = args_for_exec
@@ -390,7 +436,7 @@ class ExecutableAction(Action):
     def __repr__(self):
         return "{}({} {})".format(super().__repr__(), self.path_to_executable, " ".join(self.args_for_exec))
 
-    async def do(self, push: dict, app):  # pb: AsyncPushbullet):  # , device:Device):
+    async def on_push(self, push: dict, pb: AsyncPushbullet):
         io_loop = asyncio.get_event_loop()  # Loop handling the pushbullet IO
 
         async def _on_proc_loop():
@@ -408,7 +454,7 @@ class ExecutableAction(Action):
 
             else:
                 # Pass the incoming push via stdin (json form)
-                input_bytes = self.transform_push_to_stdin_data(push)
+                input_bytes = json.dumps(push).encode(ENCODING)
 
                 try:
                     # print("Awaiting process completion", self.path_to_executable, *self.args_for_exec)
@@ -416,9 +462,12 @@ class ExecutableAction(Action):
                         "Awaiting process completion {}".format([self.path_to_executable, *self.args_for_exec]))
                     stdout_data, stderr_data = await asyncio.wait_for(proc.communicate(input=input_bytes),
                                                                       timeout=self.timeout)
-                except asyncio.futures.TimeoutError as e:
-                    self.log.error("Execution time out after {} seconds. {}".format(self.timeout, repr(self)))
+                except asyncio.futures.TimeoutError:
+                    err_msg = "Execution timed out after {} seconds. {}".format(self.timeout, repr(self))
+                    self.log.error(err_msg)
+                    await pb.async_push_note(title="AsyncPushbullet Error", body=err_msg)
                     proc.terminate()
+
                 else:
                     # Handle the response from the subprocess
                     # This goes back on the io_loop since it may involve
@@ -427,7 +476,7 @@ class ExecutableAction(Action):
                         self.log.info("Nothing was returned from executable {}".format(
                             [self.path_to_executable, *self.args_for_exec]))
                     asyncio.run_coroutine_threadsafe(
-                        self.handle_process_response(stdout_data, stderr_data, app),
+                        self.handle_process_response(stdout_data, stderr_data, pb),
                         loop=io_loop)
                 finally:
                     self.log.debug("Process complete {}".format([self.path_to_executable, *self.args_for_exec]))
@@ -438,19 +487,12 @@ class ExecutableAction(Action):
         # a different event loop, a ProactorLoop, to handle subprocesses.
         asyncio.run_coroutine_threadsafe(_on_proc_loop(), self.proc_loop)
 
-    def transform_push_to_stdin_data(self, push: dict) -> bytes:
-        json_push = json.dumps(push)
-        input_bytes = json_push.encode(ENCODING)
-        return input_bytes
+    # def transform_push_to_stdin_data(self, push: dict) -> bytes:
+    #     json_push = json.dumps(push)
+    #     input_bytes = json_push.encode(ENCODING)
+    #     return input_bytes
 
-    async def handle_process_response(self, stdout_data: bytes, stderr_data: bytes, app):
-        """
-
-        :param bytes stdout_data:
-        :param bytes stderr_data:
-        :param ListenApp app:
-        :return:
-        """
+    async def handle_process_response(self, stdout_data: bytes, stderr_data: bytes, pb: AsyncPushbullet):
 
         # Any stderr output?
         if stderr_data != b"":
@@ -468,19 +510,23 @@ class ExecutableAction(Action):
 
         # If there's anything to respond with, send a push back
         if stderr_txt is not None:
-            title = "Error"
+            title = "AsyncPushbullet Error"
             body = "Stderr: {}\nStdout: {}".format(stderr_txt, stdout_txt)
-            await app.respond(title=title, body=body)
+            # await app.respond(title=title, body=body)
+            await pb.async_push_note(title=title, body=body)
+            del title, body
 
         elif stdout_txt is not None:
             # See if we got a structured response with JSON data
             try:
                 response = json.loads(stdout_txt)
-            except json.decoder.JSONDecodeError as e:
+            except json.decoder.JSONDecodeError:
                 # Not json, just respond with a simple push
-                title = "Response"
+                title = "AsyncPushbullet Response"
                 body = stdout_txt
-                await app.respond(title=title, body=body)
+                # await app.respond(title=title, body=body)
+                await pb.async_push_note(title=title, body=body)
+                del title, body
 
             else:
 
@@ -493,13 +539,14 @@ class ExecutableAction(Action):
                         file_type = _resp.get("file_type")
                         file_url = _resp.get("file_url")
                         file_name = _resp.get("file_name")
-                        await app.respond_file(file_name=file_name,
-                                               file_url=file_url,
-                                               file_type=file_type,
-                                               title=title,
-                                               body=body)
+                        await pb.async_push_file(file_name=file_name,
+                                                 file_url=file_url,
+                                                 file_type=file_type,
+                                                 title=title,
+                                                 body=body)
                     else:
-                        await app.respond(title=title, body=body)
+                        await pb.async_push_note(title=title, body=body)
+                    del title, body
 
                 if type(response) == list:
                     for resp in response:
@@ -508,7 +555,7 @@ class ExecutableAction(Action):
                     await _hndl_resp(response)
                 else:
                     # Not sure what was returned
-                    await app.respond(title="Response", body=str(response))
+                    await pb.async_push_note(title="Response", body=str(response))
         else:
             pass
             # Nothing sent back in stdout or stderr: send no push
@@ -533,7 +580,7 @@ class ExecutableActionSimplified(ExecutableAction):
         output_bytes = output.encode(ENCODING)
         return output_bytes
 
-    async def handle_process_response(self, stdout_data: bytes, stderr_data: bytes, app):  # pb: AsyncPushbullet):
+    async def handle_process_response(self, stdout_data: bytes, stderr_data: bytes, pb: AsyncPushbullet):
 
         if stdout_data != b"":
             data = stdout_data.decode(ENCODING, "replace")
@@ -545,44 +592,106 @@ class ExecutableActionSimplified(ExecutableAction):
                 resp = {"title": title, "body": body}
                 stdout_data = json.dumps(resp).encode(ENCODING)
 
-        await super().handle_process_response(stdout_data=stdout_data, stderr_data=stderr_data, app=app)  # pb=pb)
+        await super().handle_process_response(stdout_data=stdout_data, stderr_data=stderr_data, pb=pb)
+
+
+class ExecutableActionPython(Action):
+    """
+    Loads a python file and executes the on_push function.
+
+    async def on_push(push:dict, pb:AsyncPushbullet):
+        ...
+
+    """
+
+    def __init__(self, path: str):
+        super().__init__()
+        self.path = path  # type: str
+        self._mod_cache = None
+        self._mod_path_last_timestamp = 0.0  # type: float
+        self.module_prefix = "{}_{}".format(self.__class__.__name__, id(self))  # type: str
+
+    def __repr__(self):
+        return super().__repr__() + "({})".format(self.path)
+
+    @property
+    def module(self):
+        mtime = os.stat(self.path).st_mtime
+        mcache = self._mod_path_last_timestamp
+        if mtime > mcache and not math.isclose(mtime, mcache):
+            self.log.debug("Timestamp {} for file {} is newer than cached {}.  Reloading."
+                           .format(mtime, self.path, mcache))
+            self._mod_path_last_timestamp = mtime
+            self._mod_cache = None
+
+        if self._mod_cache is None:
+            spec = importlib.util.spec_from_file_location(self.module_prefix, self.path)
+            module = importlib.util.module_from_spec(spec)
+            spec.loader.exec_module(module)
+            self._mod_cache = module
+        return self._mod_cache
+
+    async def on_push(self, push: dict, pb: AsyncPushbullet):
+        return await self.module.on_push(push, pb)
 
 
 class ListenApp:
     def __init__(self, api_key: str,
-                 proxy: str = None,
+                 proxy=None,
                  throttle_count: int = DEFAULT_THROTTLE_COUNT,
                  throttle_seconds: float = DEFAULT_THROTTLE_SECONDS,
-                 device: str = None):
+                 device: str = None,
+                 timeout: float = None):
         self.log = logging.getLogger(__name__ + "." + self.__class__.__name__)
 
-        # self.pb = AsyncPushbullet(api_key=api_key, proxy=proxy)
-        self.account = None  # type: AsyncPushbullet
-        self._listener = None  # type: LiveStreamListener
-
+        # Passed arguments
         self.api_key = api_key
-        self.proxy = proxy
+        self.proxy = proxy  # string or Callable
         self.throttle_max_count = throttle_count
         self.throttle_max_seconds = throttle_seconds
         self.device_name = device
+        self.action_timeout = timeout
 
-        self._throttle_timestamps = []  # type: [float]
-        self.actions = []  # type: [Action]
-        # self.app_device = None  # type: Device
-        self.sent_push_idens = []  # type: List[str]
+        # Internal maintenance
+        self._account = None  # type: AsyncPushbullet
+        self._wrapped_account = None  # type: AsyncPushbullet
+        self._listener = None  # type: LiveStreamListener
+        self._throttle_timestamps = []  # type: List[float]
+        self._actions = []  # type: List[Action]
+        self._sent_push_idens = []  # type: List[str]
         self.persistent_connection = True
         self.persistent_connection_wait_interval = 10  # seconds between retry
 
+    @property
+    def account(self):
+        return self.wrapped_account
+
+    @property
+    def wrapped_account(self):
+        if self._wrapped_account is None:
+            acct = self._account
+
+            async def _new(zelf, *kargs, **kwargs):
+                resp = await zelf.__orig_async_push(*kargs, **kwargs)
+                if resp and "iden" in resp:
+                    self._sent_push_idens.append(resp.get("iden"))
+
+            acct.__orig_async_push = acct._async_push
+            acct._async_push = types.MethodType(_new, acct)
+            self._wrapped_account = acct
+
+        return self._wrapped_account
+
     def add_action(self, action: Action):
-        self.actions.append(action)
+        self._actions.append(action)
         self.log.info("Action added: {}".format(repr(action)))
 
     async def close(self):
         if self._listener is not None:
             await self._listener.close()
-        # self.pb.close_all_threadsafe()
-        if self.account is not None:
-            await self.account.async_close()
+
+        if self._account is not None:
+            await self._account.async_close()
 
     async def _throttle(self):
         """ Makes one tick and stalls if necessary """
@@ -600,90 +709,111 @@ class ListenApp:
             self.log.warning("Throttling pushes that are coming too fast. Stalling {:0.1f} seconds ...".format(stall))
             await asyncio.sleep(stall)
 
-    async def respond(self, title=None, body=None, type="note", device_iden=None):
-        """Actions can use this to respond to a push."""
-        # print("Responding with title={}, body={}".format(title, body), flush=True)
-        device = None if device_iden is None else Device(None, {"iden": device_iden})
-        resp = await self.account.async_push_note(title=title, body=body, device=device)
-        self.sent_push_idens.append(resp.get("iden"))
-
-    async def respond_file(self, file_name: str, file_url: str, file_type: str,
-                           body: str = None, title: str = None, device_iden=None):
-
-        device = None if device_iden is None else Device(None, {"iden": device_iden})
-        resp = await self.account.async_push_file(file_name=file_name,
-                                                  file_url=file_url,
-                                                  file_type=file_type,
-                                                  title=title,
-                                                  body=body,
-                                                  device=device)
-        self.sent_push_idens.append(resp.get("iden"))
-
     async def run(self):
         exit_code = 0
         while self.persistent_connection:
             try:
-                async with AsyncPushbullet(api_key=self.api_key, proxy=self.proxy) as self.account:
-                    # self.pb = AsyncPushbullet(api_key=self.api_key, proxy=self.proxy)
+                # Live update the proxy setting
+                _proxy = self.proxy() if callable(self.proxy) else self.proxy
+                if _proxy:
+                    self.log.info("Proxy: {}".format(_proxy))
+
+                print("Connecting to pushbullet...", end="", flush=True)
+                async with AsyncPushbullet(api_key=self.api_key, proxy=_proxy) as pb:
+                    self._account = pb
 
                     # If filtering on device, find or create device with that name
                     if self.device_name is not None:
-                        dev = await self.account.async_get_device(nickname=self.device_name)
+                        dev = await pb.async_get_device(nickname=self.device_name)
                         if dev is None:
-                            dev = await self.account.async_new_device(nickname=self.device_name)
+                            dev = await pb.async_new_device(nickname=self.device_name)
                             if dev is None:
                                 self.log.error("Device {} was not found and could not be created.")
                             else:
                                 self.log.info("Device {} was not found, so we created it.".format(self.device_name))
 
-                    print("Connecting to pushbullet...", end="", flush=True)
-                    async with LiveStreamListener(self.account, only_this_device_nickname=self.device_name) as pl2:
+                    async with LiveStreamListener(pb, only_this_device_nickname=self.device_name) as lsl:
                         print("Connected.", flush=True)
-                        self.log.info("Connected to Pushbullet.com service.")
-                        self._listener = pl2
+                        self.log.info("Connected to Pushbullet websocket.")
+                        self._listener = lsl
 
                         if self.device_name is None:
                             print("Awaiting pushes...")
                         else:
                             print("Awaiting pushes to device {}...".format(self.device_name))
 
-                        # print("Awaiting pushes...")
-                        async for push in pl2:
-                            self.log.info("Received push {}".format(push))
+                        async for push in lsl:
+                            self.log.info("Received push (title={}, body={}) {}"
+                                          .format(push.get("title"), push.get("body"), push))
+                            print("Received push (title={}, body={})"
+                                  .format(push.get("title"), push.get("body")))
 
                             await self._throttle()
 
-                            if push.get("iden") in self.sent_push_idens:
+                            if push.get("iden") in self._sent_push_idens:
                                 # This is one we sent - ignore it
-                                # print("Received a push we sent. Ignoring it.")
+                                self.log.debug(
+                                    "Ignoring an incoming push that we sent. (iden={})".format(push.get('iden')))
                                 continue
 
-                            for action in self.actions:  # type: Action
-                                self.log.debug("Calling action {}".format(repr(action)))
+                            async def _call_on_push(_action: Action):
+                                self.log.info("Calling action {}".format(repr(_action)))
                                 try:
-                                    await action.do(push, self)
+                                    await asyncio.wait_for(_action.on_push(push, self.wrapped_account),
+                                                           timeout=self.action_timeout)
                                     await asyncio.sleep(0)
 
+                                except asyncio.TimeoutError as te:
+                                    err_msg = "Action {} timed out after {}+ seconds".format(_action,
+                                                                                             self.action_timeout)
+                                    await pb.async_push_note(title="AsyncPushbullet Error", body=err_msg)
+                                    if not self.log.isEnabledFor(logging.DEBUG):
+                                        err_msg += " (turn on --debug to see traceback)"
+                                    self.log.warning(err_msg)
+                                    if self.log.isEnabledFor(logging.DEBUG):
+                                        traceback.print_tb(sys.exc_info()[2])
+                                    del err_msg
+
                                 except Exception as ex:
-                                    print("Action {} caused exception {}".format(action, ex))
-                                    ex.with_traceback()
+                                    err_msg = "Action {} caused exception {}".format(_action, ex)
+                                    await pb.async_push_note(title="AsyncPushbullet Error", body=err_msg)
+                                    if not self.log.isEnabledFor(logging.DEBUG):
+                                        err_msg += " (turn on --debug to see traceback)"
+                                    self.log.warning(err_msg)
+                                    if self.log.isEnabledFor(logging.DEBUG):
+                                        traceback.print_tb(sys.exc_info()[2])
+                                    del err_msg
+
+                                finally:
+                                    self.log.debug("Leaving action {}".format(repr(_action)))
+
+                            for a in self._actions:
+                                asyncio.get_event_loop().create_task(_call_on_push(a))
+
 
             except InvalidKeyError as ex:
+                print(flush=True)
                 exit_code = errors.__ERR_INVALID_API_KEY__
-                print(ex, file=sys.stderr, flush=True)
+                self.log.warning(ex)
                 self.persistent_connection = False  # Invalid key results in immediate exit
 
             except Exception as ex:
-                print(ex, file=sys.stderr, flush=True)
+                print(flush=True)
+                err_msg = "{}: {}".format(ex.__class__.__name__, ex)
+                if not self.log.isEnabledFor(logging.DEBUG):
+                    err_msg += " (turn on --debug to see traceback)"
+                self.log.warning(err_msg)
+                if self.log.isEnabledFor(logging.DEBUG):
+                    traceback.print_tb(sys.exc_info()[2])
                 exit_code = errors.__ERR_UNKNOWN__  # exit code
 
             else:
                 print("Connection closed.")
 
             finally:
-                # print("ListenApp.run() try block exiting. self.persistent_connection =", self.persistent_connection)
                 if self.persistent_connection:
-                    print("Waiting {} seconds to try again...".format(self.persistent_connection_wait_interval))
+                    print("Waiting {} seconds to try again...".format(self.persistent_connection_wait_interval),
+                          flush=True)
                     await asyncio.sleep(self.persistent_connection_wait_interval)
 
         return exit_code
